"""
Tests for geojson generation
"""

import random
import unittest

import geojson
from geojson.utils import generate_random, map_geometries


def generate_bbox():
    min_lat = random.random() * 180.0 - 90.0
    max_lat = random.random() * 180.0 - 90.0
    if min_lat > max_lat:
        min_lat, max_lat = max_lat, min_lat
    min_lon = random.random() * 360.0 - 180.0
    max_lon = random.random() * 360.0 - 180.0
    if min_lon > max_lon:
        min_lon, max_lon = max_lon, min_lon
    return [min_lon, min_lat, max_lon, max_lat]


def check_polygon_bbox(polygon, bbox):
    min_lon, min_lat, max_lon, max_lat = bbox
    eps = 1e-3
    for linear_ring in polygon['coordinates']:
        for coordinate in linear_ring:
            if not (min_lon-eps <= coordinate[0] <= max_lon+eps
                    and min_lat-eps <= coordinate[1] <= max_lat+eps):
                return False
    return True


def check_point_bbox(point, bbox):
    min_lon, min_lat, max_lon, max_lat = bbox
    eps = 1e-3
    if (
        min_lon - eps <= point["coordinates"][0] <= max_lon + eps
        and min_lat - eps <= point["coordinates"][1] <= max_lat + eps
    ):
        return True
    return False


class TestGenerateRandom(unittest.TestCase):
    def test_simple_polygon(self):
        for _ in range(5000):
            bbox = [-180.0, -90.0, 180.0, 90.0]
            result = generate_random('Polygon')
            self.assertIsInstance(result, geojson.geometry.Polygon)
            self.assertTrue(geojson.geometry.check_polygon(result))
            self.assertTrue(check_polygon_bbox(result, bbox))

    def test_bbox_polygon(self):
        for _ in range(5000):
            bbox = generate_bbox()
            result = generate_random('Polygon', boundingBox=bbox)
            self.assertIsInstance(result, geojson.geometry.Polygon)
            self.assertTrue(geojson.geometry.check_polygon(result))
            self.assertTrue(check_polygon_bbox(result, bbox))

    def test_bbox_point(self):
        for _ in range(5000):
            bbox = generate_bbox()
            result = generate_random("Point", boundingBox=bbox)
            self.assertIsInstance(result, geojson.geometry.Point)
            self.assertTrue(geojson.geometry.check_point(result))
            self.assertTrue(check_point_bbox(result, bbox))

    def test_raise_value_error(self):
        with self.assertRaises(ValueError):
            generate_random("MultiPolygon")


class TestMapGeometries(unittest.TestCase):
    def test_with_simple_type(self):
        new_point = map_geometries(
            lambda g: geojson.MultiPoint([g["coordinates"]]),
            geojson.Point((-115.81, 37.24)),
        )
        self.assertEqual(new_point, geojson.MultiPoint([(-115.81, 37.24)]))

    def test_with_feature_collection(self):
        new_point = map_geometries(
            lambda g: geojson.MultiPoint([g["coordinates"]]),
            geojson.FeatureCollection([geojson.Point((-115.81, 37.24))]),
        )
        self.assertEqual(
            new_point,
            geojson.FeatureCollection(
                [geojson.MultiPoint([geojson.Point((-115.81, 37.24))])]
            ),
        )

    def test_raise_value_error(self):
        invalid_object = geojson.Feature(type="InvalidType")
        with self.assertRaises(ValueError):
            map_geometries(lambda g: g, invalid_object)



"""Coordinate utility functions."""


def coords(obj):
    """
    Yields the coordinates from a Feature or Geometry.

    :param obj: A geometry or feature to extract the coordinates from.
    :type obj: Feature, Geometry
    :return: A generator with coordinate tuples from the geometry or feature.
    :rtype: generator
    """
    # Handle recursive case first
    if 'features' in obj:  # FeatureCollection
        for f in obj['features']:
            yield from coords(f)
    elif 'geometry' in obj:  # Feature
        yield from coords(obj['geometry'])
    elif 'geometries' in obj:  # GeometryCollection
        for g in obj['geometries']:
            yield from coords(g)
    else:
        if isinstance(obj, (tuple, list)):
            coordinates = obj
        else:
            coordinates = obj.get('coordinates', obj)
        for e in coordinates:
            if isinstance(e, (float, int)):
                yield tuple(coordinates)
                break
            for f in coords(e):
                yield f


def map_coords(func, obj):
    """
    Returns the mapped coordinates from a Geometry after applying the provided
    function to each dimension in tuples list (ie, linear scaling).

    :param func: Function to apply to individual coordinate values
    independently
    :type func: function
    :param obj: A geometry or feature to extract the coordinates from.
    :type obj: Point, LineString, MultiPoint, MultiLineString, Polygon,
    MultiPolygon
    :return: The result of applying the function to each dimension in the
    array.
    :rtype: list
    :raises ValueError: if the provided object is not GeoJSON.
    """

    def tuple_func(coord):
        return (func(coord[0]), func(coord[1]))

    return map_tuples(tuple_func, obj)


def map_tuples(func, obj):
    """
    Returns the mapped coordinates from a Geometry after applying the provided
    function to each coordinate.

    :param func: Function to apply to tuples
    :type func: function
    :param obj: A geometry or feature to extract the coordinates from.
    :type obj: Point, LineString, MultiPoint, MultiLineString, Polygon,
    MultiPolygon
    :return: The result of applying the function to each dimension in the
    array.
    :rtype: list
    :raises ValueError: if the provided object is not GeoJSON.
    """

    if obj['type'] == 'Point':
        coordinates = tuple(func(obj['coordinates']))
    elif obj['type'] in ['LineString', 'MultiPoint']:
        coordinates = [tuple(func(c)) for c in obj['coordinates']]
    elif obj['type'] in ['MultiLineString', 'Polygon']:
        coordinates = [[
            tuple(func(c)) for c in curve]
            for curve in obj['coordinates']]
    elif obj['type'] == 'MultiPolygon':
        coordinates = [[[
            tuple(func(c)) for c in curve]
            for curve in part]
            for part in obj['coordinates']]
    elif obj['type'] in ['Feature', 'FeatureCollection', 'GeometryCollection']:
        return map_geometries(lambda g: map_tuples(func, g), obj)
    else:
        raise ValueError(f"Invalid geometry object {obj!r}")
    return {'type': obj['type'], 'coordinates': coordinates}


def map_geometries(func, obj):
    """
    Returns the result of passing every geometry in the given geojson object
    through func.

    :param func: Function to apply to tuples
    :type func: function
    :param obj: A geometry or feature to extract the coordinates from.
    :type obj: GeoJSON
    :return: The result of applying the function to each geometry
    :rtype: list
    :raises ValueError: if the provided object is not geojson.
    """
    simple_types = [
        'Point',
        'LineString',
        'MultiPoint',
        'MultiLineString',
        'Polygon',
        'MultiPolygon',
    ]

    if obj['type'] in simple_types:
        return func(obj)
    elif obj['type'] == 'GeometryCollection':
        geoms = [func(geom) if geom else None for geom in obj['geometries']]
        return {'type': obj['type'], 'geometries': geoms}
    elif obj['type'] == 'Feature':
        obj['geometry'] = func(obj['geometry']) if obj['geometry'] else None
        return obj
    elif obj['type'] == 'FeatureCollection':
        feats = [map_geometries(func, feat) for feat in obj['features']]
        return {'type': obj['type'], 'features': feats}
    else:
        raise ValueError(f"Invalid GeoJSON object {obj!r}")


def generate_random(featureType, numberVertices=3,
                    boundingBox=[-180.0, -90.0, 180.0, 90.0]):
    """
    Generates random geojson features depending on the parameters
    passed through.
    The bounding box defaults to the world - [-180.0, -90.0, 180.0, 90.0].
    The number of vertices defaults to 3.

    :param featureType: A geometry type
    :type featureType: Point, LineString, Polygon
    :param numberVertices: The number vertices that a linestring or polygon
    will have
    :type numberVertices: int
    :param boundingBox: A bounding box in which features will be restricted to
    :type boundingBox: list
    :return: The resulting random geojson object or geometry collection.
    :rtype: object
    :raises ValueError: if there is no featureType provided.
    """

    from geojson import Point, LineString, Polygon
    import random
    import math

    lon_min, lat_min, lon_max, lat_max = boundingBox

    def random_lon():
        return random.uniform(lon_min, lon_max)

    def random_lat():
        return random.uniform(lat_min, lat_max)

    def create_point():
        return Point((random_lon(), random_lat()))

    def create_line():
        return LineString([create_point() for _ in range(numberVertices)])

    def create_poly():
        ave_radius = 60
        ctr_x = 0.1
        ctr_y = 0.2
        irregularity = clip(0.1, 0, 1) * math.tau / numberVertices
        spikeyness = clip(0.5, 0, 1) * ave_radius

        lower = (math.tau / numberVertices) - irregularity
        upper = (math.tau / numberVertices) + irregularity
        angle_steps = []
        for _ in range(numberVertices):
            angle_steps.append(random.uniform(lower, upper))
        sum_angle = sum(angle_steps)

        k = sum_angle / math.tau
        angle_steps = [x / k for x in angle_steps]

        points = []
        angle = random.uniform(0, math.tau)

        for angle_step in angle_steps:
            r_i = clip(random.gauss(ave_radius, spikeyness), 0, 2 * ave_radius)
            x = ctr_x + r_i * math.cos(angle)
            y = ctr_y + r_i * math.sin(angle)
            x = (x + 180.0) * (abs(lon_min - lon_max) / 360.0) + lon_min
            y = (y + 90.0) * (abs(lat_min - lat_max) / 180.0) + lat_min
            x = clip(x, lon_min, lon_max)
            y = clip(y, lat_min, lat_max)
            points.append((x, y))
            angle += angle_step

        points.append(points[0])  # append first point to the end
        return Polygon([points])

    def clip(x, min_val, max_val):
        if min_val > max_val:
            return x
        else:
            return min(max(min_val, x), max_val)

    if featureType == 'Point':
        return create_point()

    if featureType == 'LineString':
        return create_line()

    if featureType == 'Polygon':
        return create_poly()

    raise ValueError(f"featureType: {featureType} is not supported.")



from geojson.geometry import Point, LineString, Polygon
from geojson.geometry import MultiLineString, MultiPoint, MultiPolygon
from geojson.geometry import GeometryCollection
from geojson.feature import Feature, FeatureCollection
from geojson.base import GeoJSON

__all__ = ([Point, LineString, Polygon] +
           [MultiLineString, MultiPoint, MultiPolygon] +
           [GeometryCollection] +
           [Feature, FeatureCollection] +
           [GeoJSON])



from decimal import Decimal
from numbers import Number, Real

from geojson.base import GeoJSON


DEFAULT_PRECISION = 6


class Geometry(GeoJSON):
    """
    Represents an abstract base class for a WGS84 geometry.
    """

    def __init__(self, coordinates=None, validate=False, precision=None, **extra):
        """
        Initialises a Geometry object.

        :param coordinates: Coordinates of the Geometry object.
        :type coordinates: tuple or list of tuple
        :param validate: Raise exception if validation errors are present?
        :type validate: boolean
        :param precision: Number of decimal places for lat/lon coords.
        :type precision: integer
        """
        super().__init__(**extra)
        if precision is None:
            precision = DEFAULT_PRECISION
        self["coordinates"] = self.clean_coordinates(
            coordinates or [], precision)

        if validate:
            errors = self.errors()
            if errors:
                raise ValueError(f'{errors}: {coordinates}')

    @classmethod
    def clean_coordinates(cls, coords, precision):
        if isinstance(coords, cls):
            return coords['coordinates']

        new_coords = []
        if isinstance(coords, Geometry):
            coords = [coords]
        for coord in coords:
            if isinstance(coord, (list, tuple)):
                new_coords.append(cls.clean_coordinates(coord, precision))
            elif isinstance(coord, Geometry):
                new_coords.append(coord['coordinates'])
            elif isinstance(coord, (Real, Decimal)):
                new_coords.append(round(coord, precision))
            else:
                raise ValueError(f"{coord!r} is not a JSON compliant number")
        return new_coords


class GeometryCollection(GeoJSON):
    """
    Represents an abstract base class for collections of WGS84 geometries.
    """

    def __init__(self, geometries=None, **extra):
        super().__init__(**extra)
        self["geometries"] = geometries or []

    def errors(self):
        errors = [geom.errors() for geom in self['geometries']]
        return [err for err in errors if err]

    def __getitem__(self, key):
        try:
            return self.get("geometries", ())[key]
        except (KeyError, TypeError, IndexError):
            return super(GeoJSON, self).__getitem__(key)


# Marker classes.

def check_point(coord):
    if not isinstance(coord, list):
        return 'each position must be a list'
    if len(coord) not in (2, 3):
        return 'a position must have exactly 2 or 3 values'
    for number in coord:
        if not isinstance(number, Number):
            return 'a position cannot have inner positions'


class Point(Geometry):
    def errors(self):
        return check_point(self['coordinates'])


class MultiPoint(Geometry):
    def errors(self):
        return self.check_list_errors(check_point, self['coordinates'])


def check_line_string(coord):
    if not isinstance(coord, list):
        return 'each line must be a list of positions'
    if len(coord) < 2:
        return ('the "coordinates" member must be an array of '
                'two or more positions')
    for pos in coord:
        error = check_point(pos)
        if error:
            return error


class LineString(MultiPoint):
    def errors(self):
        return check_line_string(self['coordinates'])


class MultiLineString(Geometry):
    def errors(self):
        return self.check_list_errors(check_line_string, self['coordinates'])


def check_polygon(coord):
    if not isinstance(coord, list):
        return 'Each polygon must be a list of linear rings'

    if not all(isinstance(elem, list) for elem in coord):
        return "Each element of a polygon's coordinates must be a list"

    lengths = all(len(elem) >= 4 for elem in coord)
    if lengths is False:
        return 'Each linear ring must contain at least 4 positions'

    isring = all(elem[0] == elem[-1] for elem in coord)
    if isring is False:
        return 'Each linear ring must end where it started'


class Polygon(Geometry):
    def errors(self):
        return check_polygon(self['coordinates'])


class MultiPolygon(Geometry):
    def errors(self):
        return self.check_list_errors(check_polygon, self['coordinates'])


class Default:
    """
    GeoJSON default object.
    """



try:
    import simplejson as json
except ImportError:
    import json

import geojson
import geojson.factory
from geojson.mapping import to_mapping


class GeoJSONEncoder(json.JSONEncoder):

    def default(self, obj):
        return geojson.factory.GeoJSON.to_instance(obj) # NOQA


# Wrap the functions from json, providing encoder, decoders, and
# object creation hooks.
# Here the defaults are set to only permit valid JSON as per RFC 4267

def _enforce_strict_numbers(obj):
    raise ValueError(f"Number {obj!r} is not JSON compliant")


def dump(obj, fp, cls=GeoJSONEncoder, allow_nan=False, **kwargs):
    return json.dump(to_mapping(obj),
                     fp, cls=cls, allow_nan=allow_nan, **kwargs)


def dumps(obj, cls=GeoJSONEncoder, allow_nan=False, ensure_ascii=False, **kwargs):
    return json.dumps(to_mapping(obj),
                      cls=cls, allow_nan=allow_nan, ensure_ascii=ensure_ascii, **kwargs)


def load(fp,
         cls=json.JSONDecoder,
         parse_constant=_enforce_strict_numbers,
         object_hook=geojson.base.GeoJSON.to_instance,
         **kwargs):
    return json.load(fp,
                     cls=cls, object_hook=object_hook,
                     parse_constant=parse_constant,
                     **kwargs)


def loads(s,
          cls=json.JSONDecoder,
          parse_constant=_enforce_strict_numbers,
          object_hook=geojson.base.GeoJSON.to_instance,
          **kwargs):
    return json.loads(s,
                      cls=cls, object_hook=object_hook,
                      parse_constant=parse_constant,
                      **kwargs)


# Backwards compatibility
PyGFPEncoder = GeoJSONEncoder



"""
Tests for geojson/validation
"""

import unittest

import geojson

INVALID_POS = (10, 20, 30, 40)
VALID_POS = (10, 20)
VALID_POS_WITH_ELEVATION = (10, 20, 30)


class TestValidationGeometry(unittest.TestCase):

    def test_invalid_geometry_with_validate(self):
        with self.assertRaises(ValueError):
            geojson.Point(INVALID_POS, validate=True)

    def test_invalid_geometry_without_validate(self):
        geojson.Point(INVALID_POS)
        geojson.Point(INVALID_POS, validate=False)

    def test_valid_geometry(self):
        geojson.Point(VALID_POS, validate=True)

    def test_valid_geometry_with_elevation(self):
        geojson.Point(VALID_POS_WITH_ELEVATION, validate=True)

    def test_not_sequence(self):
        with self.assertRaises(ValueError) as cm:
            geojson.MultiPoint([5], validate=True)

        self.assertIn('each position must be a list', str(cm.exception))

    def test_not_number(self):
        with self.assertRaises(ValueError) as cm:
            geojson.MultiPoint([[1, '?']], validate=False)

        self.assertIn('is not a JSON compliant number', str(cm.exception))


class TestValidationGeoJSONObject(unittest.TestCase):

    def test_valid_jsonobject(self):
        point = geojson.Point((-10.52, 2.33))
        self.assertEqual(point.is_valid, True)


class TestValidationPoint(unittest.TestCase):

    def test_invalid_point(self):
        point = geojson.Point((10, 20, 30, 40))
        self.assertEqual(point.is_valid, False)

        point = geojson.Point([(10, 20), (30, 40)])
        self.assertEqual(point.is_valid, False)

    def test_valid_point(self):
        point = geojson.Point((-3.68, 40.41))
        self.assertEqual(point.is_valid, True)

    def test_valid_point_with_elevation(self):
        point = geojson.Point((-3.68, 40.41, 3.45))
        self.assertEqual(point.is_valid, True)


class TestValidationMultipoint(unittest.TestCase):

    def test_invalid_multipoint(self):
        mpoint = geojson.MultiPoint(
            [(3.5887,), (3.5887, 10.44558),
             (2.5555, 3.887, 4.56), (2.44, 3.44, 2.555, 4.56)])
        self.assertEqual(mpoint.is_valid, False)

    def test_valid_multipoint(self):
        mpoint = geojson.MultiPoint([(10, 20), (30, 40)])
        self.assertEqual(mpoint.is_valid, True)

    def test_valid_multipoint_with_elevation(self):
        mpoint = geojson.MultiPoint([(10, 20, 30), (30, 40, 50)])
        self.assertEqual(mpoint.is_valid, True)


class TestValidationLineString(unittest.TestCase):

    def test_invalid_linestring(self):
        with self.assertRaises(ValueError) as cm:
            geojson.LineString([(8.919, 44.4074)], validate=True)

        self.assertIn('must be an array of two or more positions',
                      str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            geojson.LineString([(8.919, 44.4074), [3]], validate=True)

        self.assertIn('a position must have exactly 2 or 3 values',
                      str(cm.exception))

    def test_valid_linestring(self):
        ls = geojson.LineString([(10, 5), (4, 3)])
        self.assertEqual(ls.is_valid, True)


class TestValidationMultiLineString(unittest.TestCase):

    def test_invalid_multilinestring(self):
        with self.assertRaises(ValueError) as cm:
            geojson.MultiLineString([1], validate=True)

        self.assertIn('each line must be a list of positions',
                      str(cm.exception))

        mls = geojson.MultiLineString([[(10, 5), (20, 1)], []])
        self.assertEqual(mls.is_valid, False)

    def test_valid_multilinestring(self):
        ls1 = [(3.75, 9.25), (-130.95, 1.52)]
        ls2 = [(23.15, -34.25), (-1.35, -4.65), (3.45, 77.95)]
        mls = geojson.MultiLineString([ls1, ls2])
        self.assertEqual(mls.is_valid, True)


class TestValidationPolygon(unittest.TestCase):

    def test_invalid_polygon(self):
        with self.assertRaises(ValueError) as cm:
            geojson.Polygon([1], validate=True)

        self.assertIn("Each element of a polygon's coordinates must be a list",
                      str(cm.exception))

        poly1 = geojson.Polygon(
            [[(2.38, 57.322), (23.194, -20.28), (-120.43, 19.15)]])
        self.assertEqual(poly1.is_valid, False)
        poly2 = geojson.Polygon(
            [[(2.38, 57.322), (23.194, -20.28),
                (-120.43, 19.15), (2.38, 57.323)]])
        self.assertEqual(poly2.is_valid, False)

    def test_valid_polygon(self):
        poly = geojson.Polygon(
            [[(2.38, 57.322), (23.194, -20.28),
                (-120.43, 19.15), (2.38, 57.322)]])
        self.assertEqual(poly.is_valid, True)


class TestValidationMultiPolygon(unittest.TestCase):

    def test_invalid_multipolygon(self):
        with self.assertRaises(ValueError) as cm:
            geojson.MultiPolygon([1], validate=True)

        self.assertIn("Each polygon must be a list of linear rings",
                      str(cm.exception))

        poly1 = [(2.38, 57.322), (23.194, -20.28),
                 (-120.43, 19.15), (25.44, -17.91)]
        poly2 = [(2.38, 57.322), (23.194, -20.28),
                 (-120.43, 19.15), (2.38, 57.322)]
        multipoly = geojson.MultiPolygon([poly1, poly2])
        self.assertEqual(multipoly.is_valid, False)

    def test_valid_multipolygon(self):
        poly1 = [[(2.38, 57.322), (23.194, -20.28),
                  (-120.43, 19.15), (2.38, 57.322)]]
        poly2 = [[(-5.34, 3.71), (28.74, 31.44),
                  (28.55, 19.10), (-5.34, 3.71)]]
        poly3 = [[(3.14, 23.17), (51.34, 27.14),
                  (22, -18.11), (3.14, 23.17)]]
        multipoly = geojson.MultiPolygon([poly1, poly2, poly3])
        self.assertEqual(multipoly.is_valid, True)


class TestValidationGeometryCollection(unittest.TestCase):

    def test_invalid_geometrycollection(self):
        point = geojson.Point((10, 20))
        bad_poly = geojson.Polygon([[(2.38, 57.322), (23.194, -20.28),
                                   (-120.43, 19.15), (25.44, -17.91)]])

        geom_collection = geojson.GeometryCollection(
                geometries=[point, bad_poly]
            )
        self.assertFalse(geom_collection.is_valid)

    def test_valid_geometrycollection(self):
        point = geojson.Point((10, 20))
        poly = geojson.Polygon([[(2.38, 57.322), (23.194, -20.28),
                                (-120.43, 19.15), (2.38, 57.322)]])

        geom_collection = geojson.GeometryCollection(
                geometries=[point, poly]
            )
        self.assertTrue(geom_collection.is_valid)



"""
Tests for geojson object constructor
"""

import unittest

import geojson


class TestGeoJSONConstructor(unittest.TestCase):

    def test_copy_construction(self):
        coords = [1, 2]
        pt = geojson.Point(coords)
        self.assertEqual(geojson.Point(pt), pt)

    def test_nested_constructors(self):
        a = [5, 6]
        b = [9, 10]
        c = [-5, 12]
        mp = geojson.MultiPoint([geojson.Point(a), b])
        self.assertEqual(mp.coordinates, [a, b])

        mls = geojson.MultiLineString([geojson.LineString([a, b]), [a, c]])
        self.assertEqual(mls.coordinates, [[a, b], [a, c]])

        outer = [a, b, c, a]
        poly = geojson.Polygon(geojson.MultiPoint(outer))
        other = [[1, 1], [1, 2], [2, 1], [1, 1]]
        poly2 = geojson.Polygon([outer, other])
        self.assertEqual(geojson.MultiPolygon([poly, poly2]).coordinates,
                         [[outer], [outer, other]])



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'geojson'
copyright = '2024, Sean Gillies, Matthew Russell, Corey Farwell, Blake Grotewold, \
    Zsolt Ero, Sergey Romaov, Ray Riga'
author = 'Sean Gillies, Matthew Russell, Corey Farwell, Blake Grotewold, Zsolt Ero, \
    Sergey Romaov, Ray Riga'
release = '3.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinxcontrib.jquery']

templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_theme_path = ['_theme']
html_static_path = ['_static']



from io import StringIO
import unittest

import geojson


class FeaturesTest(unittest.TestCase):
    def test_protocol(self):
        """
        A dictionary can satisfy the protocol
        """
        f = {
            'type': 'Feature',
            'id': '1',
            'geometry': {'type': 'Point', 'coordinates': [53.0, -4.0]},
            'properties': {'title': 'Dict 1'},
        }

        json = geojson.dumps(f, sort_keys=True)
        self.assertEqual(json, '{"geometry":'
                               ' {"coordinates": [53.0, -4.0],'
                               ' "type": "Point"},'
                               ' "id": "1",'
                               ' "properties": {"title": "Dict 1"},'
                               ' "type": "Feature"}')

        o = geojson.loads(json)
        output = geojson.dumps(o, sort_keys=True)
        self.assertEqual(output, '{"geometry":'
                                 ' {"coordinates": [53.0, -4.0],'
                                 ' "type": "Point"},'
                                 ' "id": "1",'
                                 ' "properties": {"title": "Dict 1"},'
                                 ' "type": "Feature"}')

    def test_unicode_properties(self):
        with open("tests/data.geojson") as file_:
            obj = geojson.load(file_)
        geojson.dump(obj, StringIO())

    def test_feature_class(self):
        """
        Test the Feature class
        """

        from geojson.examples import SimpleWebFeature
        feature = SimpleWebFeature(
            id='1',
            geometry={'type': 'Point', 'coordinates': [53.0, -4.0]},
            title='Feature 1', summary='The first feature',
            link='http://example.org/features/1'
        )

        # It satisfies the feature protocol
        self.assertEqual(feature.id, '1')
        self.assertEqual(feature.properties['title'], 'Feature 1')
        self.assertEqual(feature.properties['summary'], 'The first feature')
        self.assertEqual(feature.properties['link'],
                         'http://example.org/features/1')
        self.assertEqual(geojson.dumps(feature.geometry, sort_keys=True),
                         '{"coordinates": [53.0, -4.0], "type": "Point"}')

        # Encoding
        json = ('{"geometry": {"coordinates": [53.0, -4.0],'
                ' "type": "Point"},'
                ' "id": "1",'
                ' "properties":'
                ' {"link": "http://example.org/features/1",'
                ' "summary": "The first feature",'
                ' "title": "Feature 1"},'
                ' "type": "Feature"}')
        self.assertEqual(geojson.dumps(feature, sort_keys=True), json)

        # Decoding
        factory = geojson.examples.create_simple_web_feature
        json = ('{"geometry": {"type": "Point",'
                ' "coordinates": [53.0, -4.0]},'
                ' "id": "1",'
                ' "properties": {"summary": "The first feature",'
                ' "link": "http://example.org/features/1",'
                ' "title": "Feature 1"}}')
        feature = geojson.loads(json, object_hook=factory)
        self.assertEqual(repr(type(feature)),
                         "<class 'geojson.examples.SimpleWebFeature'>")
        self.assertEqual(feature.id, '1')
        self.assertEqual(feature.properties['title'], 'Feature 1')
        self.assertEqual(feature.properties['summary'], 'The first feature')
        self.assertEqual(feature.properties['link'],
                         'http://example.org/features/1')
        self.assertEqual(geojson.dumps(feature.geometry, sort_keys=True),
                         '{"coordinates": [53.0, -4.0], "type": "Point"}')

    def test_geo_interface(self):
        class Thingy:
            def __init__(self, id, title, x, y):
                self.id = id
                self.title = title
                self.x = x
                self.y = y

            @property
            def __geo_interface__(self):
                return ({"id": self.id,
                         "properties": {"title": self.title},
                         "geometry": {"type": "Point",
                                      "coordinates": (self.x, self.y)}})

        ob = Thingy('1', 'thingy one', -106.0, 40.0)
        self.assertEqual(geojson.dumps(ob.__geo_interface__['geometry'],
                                       sort_keys=True),
                         '{"coordinates": [-106.0, 40.0], "type": "Point"}')
        self.assertEqual(geojson.dumps(ob, sort_keys=True),
                         ('{"geometry": {"coordinates": [-106.0, 40.0],'
                          ' "type": "Point"},'
                          ' "id": "1",'
                          ' "properties": {"title": "thingy one"}}'))



from setuptools import setup
import sys
import re


with open("README.rst") as readme_file:
    readme_text = readme_file.read()

VERSIONFILE = "geojson/_version.py"
verstrline = open(VERSIONFILE).read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError(f"Unable to find version string in {VERSIONFILE}.")


major_version, minor_version = sys.version_info[:2]
if not (major_version == 3 and 7 <= minor_version <= 13):
    sys.stderr.write("Sorry, only Python 3.7 - 3.13 are "
                     "supported at this time.\n")
    exit(1)

setup(
    name="geojson",
    version=verstr,
    description="Python bindings and utilities for GeoJSON",
    license="BSD",
    keywords="gis geography json",
    author="Sean Gillies",
    author_email="sgillies@frii.com",
    maintainer="Ray Riga",
    maintainer_email="ray.maintainer@gmail.com",
    url="https://github.com/jazzband/geojson",
    long_description=readme_text,
    packages=["geojson"],
    package_dir={"geojson": "geojson"},
    package_data={"geojson": ["*.rst"]},
    install_requires=[],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering :: GIS",
    ]
)
